/* medroot -- compute median root image */

/* contains: main(), GetArgsMR(), ImgDiff() */

/* dependencies: ReadRasterFile(), MorphSub(), WriteRasterFile()   */

/* mclean version 1.0   15 July 1991               */
/* morphological image noise reduction program     */
/*                                                 */
/* by Richard Alan Peters II                       */
/* Department of Electrical Engineering            */
/* Vanderbilt University School of Engineering     */
/* Nashville, TN 37235                             */
/* rap2@vuse.vanderbilt.edu                        */
/*                                                 */
/* This software is freely redistributable if      */
/* the author's name and affiliation are included. */

#include "math.h"
#include "values.h"
#include "morph_sub.h"
#include "rasterio.h"

#define FNLGTH 64   /* maximum file name string length */

static void GetArgsMR();
static void OptErr();
static int  ImgDiff();

static struct rasterfile RasHd;    /* raster file header */
static unsigned char *cmap;        /* colormap */
 
main( argc, argv ) 
   unsigned int argc;    /* count of arguments parsed by operating system    */
   char *argv[];         /* pointers to arguments parsed by operating system */
   {
   char *SEname;  /* structuring element (SE) path name */
   int autoSE;    /* T => create a disk-shaped SE locally */
   int sX,sY;     /* SE horizontal, vertical dimensions */
   int sZ;        /* largest gray level in SE */
   int SEType;    /* binary or gray-level SE */
   int ImgType;   /* gray-level or binary */
   int LThresh;   /* lower binary threshold value */
   int UThresh;   /* upper binary threshold value */
   int Rank;      /* rank for rank filter */
   int diff;      /* sum of absolute differences between images */


   byte *Img;          /* input image  */
   byte *Out;          /* output image */
   byte *I,*S,*T;      /* image pointers */
   int X,Y;            /* image horiz, vert. dimensions  */
   int M;              /* number of pixels in image      */
   int i;              /* index variable    */
   int max;            /* maximum number of iterations allowed */
   float L1norm;       /* L1 norm or original image */
   float NormDiff;     /* normalized difference */
   float NormDiffP;    /* previous normalized difference */
   float Tolerance;    /* smallest acceptible change in normalized difference */
   FILE *fp;           /* file pointer  */
   char fname[FNLGTH]; /* name of output file for smooth input/output */



   /* begin */

   if (argc < 5 ) /* display usage */
      {
      fprintf(stderr,
         "usage: medroot  InFile  OutFile  [-m nnn (max # iterations)] \n\
      [-d (smallest acceptible difference)] [-i g|b]  [-s g|b]  [-t lll [uuu]]\n\
       -k SEFile | 3x3 | 5x5 | plus | auto xxx yyy [zzz]\n");
      exit(0);
      }

   /* get user arguments */

   GetArgsMR(argc,argv,&SEname,&autoSE,&sX,&sY,&sZ,&SEType,
           &ImgType,&Rank,&LThresh,&UThresh,&max,&Tolerance);

   
   /* read the input image into memory */
   
   if ( !(fp = OpenFile( argv[1], NULL, "r" )) ) exit( 0 );
   if ( ReadRasterFile( fp, &RasHd, &cmap, &Img, 0, 0, 0, 0, ALLOC ) ) 
      exit(0);
   fclose( fp );

 
   /* remap the image */
   ExtractLuminance( &RasHd, cmap, Img, 0, 0, 0, 0 );


  /* get the image dimensions */

   X = RasHd.ras_width;
   Y = RasHd.ras_height;
   M = X*Y;


   /* allocate space for output image */

   if ( !(Out = (byte *)calloc( M, sizeof(byte) )) ) 
      {
      fprintf(stderr,"main: Unable to allocate output image.\n");
      exit(0);
      }

   L1norm = (float)ImgDiff( Img, Out, X, Y );
   NormDiff  = MAXFLOAT / 2;
   NormDiffP = MAXFLOAT;


   Rank = 0;
   diff = 1;
   i = 1;
   while ( diff && ( (!max) || (i <= max) ) 
                && (NormDiff >= Tolerance) && (NormDiffP > NormDiff) )
      {
      fprintf( stderr, "median root: iteration %d.  ",i);
      fflush ( stderr );

      /* do the morph op */
      MorphSub( RANK, Img, Out, X, Y, ImgType, FALSE, LThresh, UThresh,
                SEname, autoSE, sX, sY, sZ, SEType, SET, Rank, TRUE, FALSE);

      diff = ImgDiff( Img, Out, X, Y );

      NormDiffP = NormDiff;
      NormDiff = (float)diff / L1norm;
      fprintf( stderr, "Normalized L1 difference: %e\n",NormDiff);

      memcpy( Img, Out, M );
      ++i;
      }

   /* create output file */
   if ( !(fp = OpenFile( argv[2], NULL, "w" )) ) exit( 0 );
   if ( WriteRasterFile( fp, &RasHd, cmap, Out, 0, 0, 0, 0 ) ) 
      exit(0);
   fclose( fp );

   free( Img );
   free( Out );
   if ( cmap ) free( cmap );

   } /* end */




static int ImgDiff ( I1, I2, X, Y )
   byte *I1;  /* image 1 */
   byte *I2;  /* image 2 */
   int X,Y;   /* image horizontal, vertical dimensions */
   {
   byte *A = I1;
   byte *B = I2;
   int   n = X*Y;
   int accum;
   int i;

   accum = 0;
   for (i=n; i; --i)
      {
       accum += abs( (int)*A - (int)*B );
      ++A;
      ++B;
      }

   return(accum);
   }



/* parse (argc, argv) argument list for morph */

static void GetArgsMR(argc,argv,sname,autoSE,x,y,z,
            SEType,ImgType,rank,lthresh,uthresh,maxiter,delta) 
   unsigned int argc;   /* count of arguments parsed by operating system    */
   char *argv[];        /* pointers to arguments parsed by operating system */
   char **sname;        /* structuring element (SE) path name */
   int *autoSE;         /* use canned or auto generated SE */
   int *x,*y,*z;        /* SE horizontal, vertical dimensions, gray-level */
   int *SEType;         /* binary or gray-level SE */
   int *ImgType;        /* gray-level or binary */
   int *rank;           /* rank for rank filter */
   int *lthresh;        /* lower binary threshold value */
   int *uthresh;        /* upper binary threshold value */
   int *maxiter;        /* maximum number of iterations permited */
   float *delta;        /* minimum acceptible normalized difference */
   {
   int n = argc;
   int i;

   /* zero everything initially */
   *autoSE=0;
   *x=0;
   *y=0;
   *z=0;
   *SEType=0;
   *ImgType=0;
   *rank=0;
   *lthresh=0;
   *uthresh=0;
   *maxiter=0;
   *delta=0.0;


   /* get arguments */
   while ( --n )
      {
      if ( argv[n][0] == '-' )          /* then this is a switch */
         {
         if ( argv[n][1] == 'k' )       /* SE file name follows */
            {
            if ( !strcasecmp( argv[n+1], "3x3" ) )
               {
               *autoSE = S3X3;
               *sname = NULL;
               }
            else if ( !strcasecmp( argv[n+1], "plus" ) )
               {
               *autoSE = PLUS;
               *sname = NULL;
               }
            else if ( !strcasecmp( argv[n+1], "5x5" ) )
               {
               *autoSE = S5X5;
               *sname = NULL;
               }
            else if ( !strcasecmp( argv[n+1], "auto" ) )
               {
               *autoSE = AUTO;
               *x = atoi( argv[n+2] );
               *y = atoi( argv[n+3] );
               if ( ((n+4) < argc) && isdigit( argv[n+4][0] ) )
                  *z = atoi( argv[n+4] );
               /* note if (*autoSE != 0), setupop sets SEType = GRASE */
               *sname = NULL;
               }
            else
               *sname = argv[n+1];
            }
         else if ( argv[n][1] == 's' )  /* SE type follows */
            {
            switch (argv[n+1][0])
               {
               case 'b' : { *SEType = BINSE; break; }
               case 'g' : { *SEType = GRASE; break; }
               default  : { OptErr( argv[n], argv[n+1] ); }
               }
            }
         else if ( argv[n][1] == 'i' )  /* image type follows */
            {
            switch (argv[n+1][0])
               {
               case 'g' : { *ImgType = GRAIMG; break; }
               case 'b' : { *ImgType = BINIMG; break; }
               default  : { OptErr( argv[n], argv[n+1] ); }
               }
            }
         else if ( argv[n][1] == 't' )  /* threshold value(s) follows */
            {
            *lthresh = atoi( argv[n+1] );
            if ( n+2 < argc  &&  isdigit( argv[n+2][0] ) )
                *uthresh = atoi( argv[n+2] );
            else
                *uthresh = WHITE;
            if (*lthresh > *uthresh) 
                OptErr( argv[n], "lthresh > uthresh" );
            } 
         else if ( argv[n][1] == 'm' )  /* maximum iterations follows */
            {
            if ( n+1 < argc  &&  isdigit( argv[n+1][0] ) )
                *maxiter = atoi( argv[n+1] );
            else
                OptErr( argv[n], "number not specified" );
            } 
         else if ( argv[n][1] == 'd' )  /* minimum difference follows */
            {
            if ( n+1 < argc  &&  isdigit( argv[n+1][0] ) )
                *delta = atof( argv[n+1] );
            else
                OptErr( argv[n], "number not specified" );
            } 
         }    /* end of if ( argv[n][0] == '-' ) */
      }   /* end of  while ( --n ) "get arguments" loop */

   return;
   }
   


static void OptErr( s, t )
   char *s,*t;
   {
   fprintf(stderr,"Invalid Op: %s %s\n",s,t);
   exit( 0 );
   }


