/* recon -- performs grayscale reconstruction of an image from a smoothed   */
/*          version. Works in an 8-connected or 4-connected neighborhood.  */
/*                                                                          */
/* This is an implementation of the fast hybrid grayscale reconstruction    */
/* algorithm described in Vincent, L., "Morphological reconstruction in     */
/* image analysis: applications and efficient algorithms, IEEE Trans. Image */
/* Processing, Vol. 2, No. 2, April 1993.                                   */
/*                                                                          */
/* I have modified the algorithm so that (1) the smooth image need not be   */
/* everywhere less than (or greater than) the original image. (2) it uses   */
/* either an 8-connected or a 4-connected neighborhood.                     */

/* version 2.0   2 March 1994  */

/* by Richard Alan Peters II                       */
/* Department of Electrical Engineering            */
/* Vanderbilt University School of Engineering     */ 
/* Nashville, TN 37235                             */ 
/* rap2@vuse.vanderbilt.edu                        */ 
/*                                                 */ 
/* This software is freely redistributable if      */ 
/* the author's name and affiliation are included. */


#include <math.h>        
#include <stdio.h>
#include <string.h>
#include <values.h>
#include "rasterio.h"
#include "recon.h"


main( argc, argv ) 
   unsigned int argc;    /* count of arguments parsed by operating system    */
   char *argv[];         /* pointers to arguments parsed by operating system */
   {
   FILE *fp;                /* file pointer */
   struct rasterfile HI,HS; /* rasterfile headers */
   byte *CMapI,*CMapS;      /* colormaps */
   byte *I,*S,*R;           /* image pointers */
   int X,Y,N;               /* image X,Y dimensions and size, N */
   int nbhd;                /* either 4 or 8 - connected */


   if (argc < 4)
      {
      fprintf(stderr,
         "usage: recon  [-8|4]  original_image  smooth_image  reconstructed_image\n");
      exit(0);
      }


   /* default neighborhood is 8-connected */
   nbhd = 8;


   /* get optional argument */
   if ( argv[1][0] == '-' )
      {
      nbhd = atoi( argv[1]+1 );
      if ( !(nbhd == 8  || nbhd == 4) )
         {
         fprintf(stderr,"neighborhood must be 8-connected (-8) or 4-connected (-4).\n");
         exit( 0 );
         } 
      }


   /* open original image file */
   if ( !(fp = OpenFile( argv[argc-3], NULL, "r" )) ) exit( 0 );
   if (  ReadRasterFile( fp, &HI, &CMapI, &I, 0, 0, 0, 0, ALLOC )  ) exit( 0 );
   fclose( fp );


   /* remap the image */
   ExtractLuminance( &HI, CMapI, I, 0, 0, 0, 0 );


   /* image dimensions */
   X = HI.ras_width;
   Y = HI.ras_height;
   N = X*Y;


   /* open smooth (marker) image  */
   if ( !(fp = OpenFile( argv[argc-2], NULL, "r" )) ) exit( 0 );
   if (  ReadRasterFile( fp, &HS, &CMapS, &S, 0, 0, 0, 0, ALLOC )  ) exit( 0 );
   fclose( fp );


   /* check dimensions */
   if ( (HS.ras_width != X)  ||  (HS.ras_height != Y) )
      {
      fprintf(stderr, "The two input images have unequal dimensions\n");
      exit( 0 );
      }


   /* remap the image */
   ExtractLuminance( &HS, CMapS, S, 0, 0, 0, 0 );


   /* allocate space for the reconstructed image */
   if ( !(R = (byte *)calloc( N, sizeof(byte) )) )
      {
      fprintf(stderr, "Unable to allocate the reconstructed image.");
      exit( 0 );
      }


   /* do the reconstruction */
   ReconSub( I, S, R, X, Y, nbhd, TRUE ); 
  

   /* create output file */
   if ( !(fp = OpenFile( argv[argc-1], NULL, "w" )) ) exit( 0 );

   WriteRasterFile( fp, HI, CMapI, R, 0, 0, 0, 0 ); 
               
   fclose( fp );
   }

