/* recon_sub.c --Vincent reconstruction subroutine */

/* contains: ReconSub(), ChkTopBot(), ReconTop(), ReconBot()               */
/*           ForScanTop8(), BackScanTop8(), ForScanBot8(), BackScanBot8(), */ 
/*           ForScanTop4(), BackScanTop4(), ForScanBot4(), BackScanBot4(), */
/*           MakeTopBot(), MakeRecon(), Difference()                       */

/* dependencies:  *OpenFile(), ReadRasterFile(), WriteRasterFile(),        */
/*                 ExtractLuminance();                                     */

/* This is an implementation of the fast hybrid grayscale reconstruction    */
/* algorithm described in Vincent, L., "Morphological reconstruction in     */
/* image analysis: applications and efficient algorithms, IEEE Trans. Image */
/* Processing, Vol. 2, No. 2, April 1993.                                   */
/*                                                                          */
/* I have modified the algorithm so that (1) the smooth image need not be   */
/* everywhere less than (or greater than) the original image. (2) it uses   */
/* either an 8-connected or a 4-connected neighborhood.                     */
/* 

/* version 2.0  2 March 1994  */

/* by Richard Alan Peters II                       */
/* Department of Electrical Engineering            */
/* Vanderbilt University School of Engineering     */ 
/* Nashville, TN 37235                             */ 
/* rap2@vuse.vanderbilt.edu                        */ 
/*                                                 */ 
/* This software is freely redistributable if      */ 
/* the author's name and affiliation are included. */

#include <stdio.h>
#include <math.h>        
#include <values.h>
#include "recon.h"


void ReconSub( Img, Sm, Rec, X, Y, Nbhd, verbose )
   byte *Img;  /* object (original) image */
   byte *Sm;   /* smooth (marker) image   */
   byte *Rec;  /* reconstructed image     */
   int   X, Y; /* image dimensions        */
   int   Nbhd; /* neighborhood type (8-connected or 4-connected) */
   int   verbose; /* T => announce iterations. F => don't. */
   {
   byte *I,*T,*B;           /* image pointers */
   byte *J,*J1,*J2,*JB,*JT; /* image pointers */
   byte *Tmp;               /* image pointer */
   int   N;                 /* image X,Y dimensions and size, N */
   int bot,top;             /* flags */


   /* compute size in byte of image */
   N = X*Y;


   /* check arguments */
   if ( !( Img  &&  Sm  &&  Rec ) ) 
      {
      fprintf(stderr,"ReconSub: Null image buffer received.\n");
      exit( 0 );
      }
   if ( !(Nbhd == 8  ||  Nbhd == 4) )
      {
      fprintf(stderr,"ReconSub: Nbhd = %d (it must be either 4 or 8).\n",Nbhd);
      exit( 0 );
      }
   if ( !N )
      {
      fprintf(stderr,"ReconSub: Zero image dimension.\n",Nbhd);
      exit( 0 );
      }


   /* make a working copy of the smooth image */
   if ( !(Tmp = (byte *)calloc( N, sizeof(byte) )) )
      {
      fprintf(stderr,"ReconSub: Unable to allocate the temporary image.");
      exit( 0 );
      }
   memcpy( Tmp, Sm, N );


   /* set up pointers */
   I = Img;
   J = Tmp;


   /* check to see if top, bot, or both are necessary */
   ChkTopBot( I, J, N, &top, &bot );


   /* recon procs */
   if ( top && bot )
      {
      /* allocate the top, bot, and secondary images */
      if ( !(T = (byte *)calloc( N, sizeof(byte) )) )
         {
         fprintf(stderr,"ReconSub: Unable to allocate the top image.");
         exit( 0 );
         }
      if ( !(B = (byte *)calloc( N, sizeof(byte) )) )
         {
         fprintf(stderr,"ReconSub: Unable to allocate the bot image.");
         exit( 0 );
         }
      /* make the secondary J arrays */
      if ( !(J1 = (byte *)calloc( N, sizeof(byte) )) )
         {
         fprintf(stderr,"ReconSub: Unable to allocate the J1 array.");
         exit( 0 );
         }
      if ( !(J2 = (byte *)calloc( N, sizeof(byte) )) )
         {
         fprintf(stderr,"ReconSub: Unable to allocate the J2 array.");
         exit( 0 );
         }

      /* make the top and bot, T and B, from the orig, I, and smooth J */
      MakeTopBot( I, J, T, B, N);

      /* copy the smooth (marker) image J to J2 */
      memcpy( J2, J, N );

      /* reconstruct top from (and in) J using J1 as scratch space */
      ReconTop( T, J, J1, X, Y, Nbhd, verbose );

      /* copy reconstruction to T */
      memcpy( T, J, N );

      /* reconstruct bot from (and in) J2 using J1 as scratch space */
      ReconBot( B, J2, J1, X, Y, Nbhd, verbose );

      /* copy recon to B */
      memcpy( B, J2, N );

      /* make result */
      MakeRecon( I, J, T, B, N);
      }
   else if ( top )
      {
      /* make the secondary J array */
      if ( !(JT = (byte *)calloc( N, sizeof(byte) )) )
         {
         fprintf(stderr,"ReconSub: Unable to allocate the secondary J array.");
         exit( 0 );
         }
      /* reconstruct top from (and in) J using JT as scratch space */
      ReconTop( I, J, JT, X, Y, Nbhd, verbose );
      }
   else if ( bot )
      {
      /* make the secondary J array */
      if ( !(JB = (byte *)calloc( N, sizeof(byte) )) )
         {
         fprintf(stderr,"ReconSub: Unable to allocate the secondary J array.");
         exit( 0 );
         }
      /* reconstruct top from (and in) J using JB as scratch space */
      ReconBot( I, J, JB, X, Y, Nbhd, verbose );
      }
   else 
      {
      fprintf( stderr,"Warning: the input and marker images are identical\n" );
      /* copy the original image to the reconstructed and return */
      memcpy( Rec, Img, N );
      return;
      }

   /* Copy reconstructed image to Rec  */
   memcpy( Rec, J, N );

   free ( Tmp );
   if (  T ) free(  T );
   if (  B ) free(  B );
   if ( J1 ) free( J1 );
   if ( J2 ) free( J2 );
   if ( JT ) free( JT );
   if ( JB ) free( JB );
   return;
   }



void ChkTopBot( I, J, N, top, bot )
   byte *I;
   byte *J;
   int N;
   int *top,*bot;
   {
   byte *u,*v;
   int i;

  *top = FALSE;
   u=I;
   v=J;
   for (i=N; i; --i )
      {
      if (*u > *v)
         {
         *top = TRUE;
         break;
         }
      else
         {
         ++u;
         ++v;
         }
      }

  *bot = FALSE;
   u=I;
   v=J;
   for (i=N; i; --i )
      {
      if (*u < *v)
         {
         *bot = TRUE;
         break;
         }
      else
         {
         ++u;
         ++v;
         }
      }
  return;
   }



void ReconTop( I, J, J2, X, Y, nbhd, verbose)
   byte *I;
   byte *J;
   byte *J2;
   int X,Y;
   int nbhd;
   int verbose;
   {
   int i,d,N;
   
   N = X*Y;
   /* main reconstruction loop */
   if ( verbose ) fprintf(stderr,"reconstructing top\n");
   i=0;
   do
      {
      ++i;

      /* announce iteration */
      if ( verbose ) fprintf(stderr,"top iteration %d\n",i );

      /* save latest J */
      memcpy( J2, J, N );

      if ( nbhd == 8 )
         {
         /* scan forward */
         ForScanTop8( I, J, X, Y );
 
         /* scan back */
         BackScanTop8( I, J, X, Y );
         }
      else
         {
         /* scan forward */
         ForScanTop4( I, J, X, Y );
 
         /* scan back */
         BackScanTop4( I, J, X, Y );
         }

      /* compute difference */
      d = Difference( J, J2, N );
      }
   while ( d );
   return;
   }
   



void ReconBot( I, J, J2, X, Y, nbhd, verbose)
   byte *I;
   byte *J;
   byte *J2;
   int X,Y;
   int nbhd;
   int verbose;
   {
   int i,d,N;
   
   N = X*Y;
   /* main reconstruction loop */
   if ( verbose ) fprintf(stderr,"reconstructing bot\n");
   i=0;
   do
      {
      ++i;

      /* announce iteration */
      if ( verbose ) fprintf(stderr,"bot iteration %d\n",i );

      /* save latest J */
      memcpy( J2, J, N );

      if ( nbhd == 8 )
         {
         /* scan forward */
         ForScanBot8( I, J, X, Y );
 
         /* scan back */
         BackScanBot8( I, J, X, Y );
         }
      else
         {
         /* scan forward */
         ForScanBot4( I, J, X, Y );
 
         /* scan back */
         BackScanBot4( I, J, X, Y );
         }

      /* compute difference */
      d = Difference( J, J2, N );
      }
   while ( d );
   return;
   }
   



void ForScanTop8( I, J, X, Y )
   byte *I;
   byte *J;
   int X,Y;
   {
   byte *a,*b,*c,*d,*p,*q;
   byte m;
   int i,j;
   int r,s;
   
   r = 0;
   for ( j=Y-1; j; --j )
      {
      s  = r;     /* pixel (i-1,j-1) loc */
      r += X;     /* pixel above (i-1,j) loc */
      a  = J+s;   /* J(i-1,j-1) */
      b  = a+1;   /* J(i  ,j-1) */
      c  = b+1;   /* J(i+1,j-1) */
      d  = J+r;   /* J(i-1,j  ) */
      p  = d+1;   /* J(i  ,j  ) */
      q  = I+r+1; /* I(i  ,j  ) */
      for ( i=X-2; i; --i )
         {
          m = *a>*b ? *a : *b;
          m =  m>*c ?  m : *c;
          m =  m>*d ?  m : *d;
          m =  m>*p ?  m : *p;
         *p =  m<*q ?  m : *q;
         ++a;
         ++b;
         ++c;
         ++d;
         ++p;
         ++q;
         }
      }
   return;
   }


void BackScanTop8( I, J, X, Y )
   byte *I;
   byte *J;
   int X,Y;
   {
   byte *a,*b,*c,*d,*p,*q;
   byte m;
   int i,j;
   int r,s;
   
   r = (X*Y)-1;
   for ( j=Y-1; j; --j )
      {
      s  = r;     /* pixel (i+1,j+1) loc */
      r -= X;     /* pixel above (i+1,j) loc */
      a  = J+s;   /* J(i+1,j+1) */  
      b  = a-1;   /* J(i  ,j+1) */
      c  = b-1;   /* J(i-1,j+1) */
      d  = J+r;   /* J(i+1,j  ) */ 
      p  = d-1;   /* J(i  ,j  ) */
      q  = I+r-1; /* I(i  ,j  ) */
      for ( i=X-2; i; --i )
         {
          m = *a>*b ? *a : *b;
          m =  m>*c ?  m : *c;
          m =  m>*d ?  m : *d;
          m =  m>*p ?  m : *p;
         *p =  m<*q ?  m : *q;
         --a;
         --b;
         --c;
         --d;
         --p;
         --q;
         }
      }
   return;
   }



void ForScanBot8( I, J, X, Y )
   byte *I;
   byte *J;
   int X,Y;
   {
   byte *a,*b,*c,*d,*p,*q;
   byte m;
   int i,j;
   int r,s;
   
   r = 0;
   for ( j=Y-1; j; --j )
      {
      s  = r;     /* pixel (i-1,j-1) loc */
      r += X;     /* pixel above (i-1,j) loc */
      a  = J+s;   /* J(i-1,j-1) */
      b  = a+1;   /* J(i  ,j-1) */
      c  = b+1;   /* J(i+1,j-1) */
      d  = J+r;   /* J(i-1,j  ) */
      p  = d+1;   /* J(i  ,j  ) */
      q  = I+r+1; /* I(i  ,j  ) */
      for ( i=X-2; i; --i )
         {
          m = *a<*b ? *a : *b;
          m =  m<*c ?  m : *c;
          m =  m<*d ?  m : *d;
          m =  m<*p ?  m : *p;
         *p =  m>*q ?  m : *q;
         ++a;
         ++b;
         ++c;
         ++d;
         ++p;
         ++q;
         }
      }
   return;
   }


void BackScanBot8( I, J, X, Y )
   byte *I;
   byte *J;
   int X,Y;
   {
   byte *a,*b,*c,*d,*p,*q;
   byte m;
   int i,j;
   int r,s;
   
   r = (X*Y)-1;
   for ( j=Y-1; j; --j )
      {
      s  = r;     /* pixel (i+1,j+1) loc */
      r -= X;     /* pixel above (i+1,j) loc */
      a  = J+s;   /* J(i+1,j+1) */  
      b  = a-1;   /* J(i  ,j+1) */
      c  = b-1;   /* J(i-1,j+1) */
      d  = J+r;   /* J(i+1,j  ) */ 
      p  = d-1;   /* J(i  ,j  ) */
      q  = I+r-1; /* I(i  ,j  ) */
      for ( i=X-2; i; --i )
         {
          m = *a<*b ? *a : *b;
          m =  m<*c ?  m : *c;
          m =  m<*d ?  m : *d;
          m =  m<*p ?  m : *p;
         *p =  m>*q ?  m : *q;
         --a;
         --b;
         --c;
         --d;
         --p;
         --q;
         }
      }
   return;
   }
   



void ForScanTop4( I, J, X, Y )
   byte *I;
   byte *J;
   int X,Y;
   {
   byte *b,*d,*p,*q;
   byte m;
   int i,j;
   int r,s;
   
   r = 0;
   for ( j=Y-1; j; --j )
      {
      s  = r;     /* pixel (i-1,j-1) loc */
      r += X;     /* pixel above (i-1,j) loc */
      b  = J+s+1; /* J(i  ,j-1) */
      d  = J+r;   /* J(i-1,j  ) */
      p  = d+1;   /* J(i  ,j  ) */
      q  = I+r+1; /* I(i  ,j  ) */
      for ( i=X-2; i; --i )
         {
          m = *b>*d ? *b : *d;
          m =  m>*p ?  m : *p;
         *p =  m<*q ?  m : *q;
         ++b;
         ++d;
         ++p;
         ++q;
         }
      }
   return;
   }


void BackScanTop4( I, J, X, Y )
   byte *I;
   byte *J;
   int X,Y;
   {
   byte *b,*d,*p,*q;
   byte m;
   int i,j;
   int r,s;
   
   r = (X*Y)-1;
   for ( j=Y-1; j; --j )
      {
      s  = r;     /* pixel (i+1,j+1) loc */
      r -= X;     /* pixel above (i+1,j) loc */
      b  = J+s-1;   /* J(i  ,j+1) */
      d  = J+r;   /* J(i+1,j  ) */ 
      p  = d-1;   /* J(i  ,j  ) */
      q  = I+r-1; /* I(i  ,j  ) */
      for ( i=X-2; i; --i )
         {
          m = *b>*d ? *b : *d;
          m =  m>*p ?  m : *p;
         *p =  m<*q ?  m : *q;
         --b;
         --d;
         --p;
         --q;
         }
      }
   return;
   }



void ForScanBot4( I, J, X, Y )
   byte *I;
   byte *J;
   int X,Y;
   {
   byte *b,*d,*p,*q;
   byte m;
   int i,j;
   int r,s;
   
   r = 0;
   for ( j=Y-1; j; --j )
      {
      s  = r;     /* pixel (i-1,j-1) loc */
      r += X;     /* pixel above (i-1,j) loc */
      b  = J+s+1; /* J(i  ,j-1) */
      d  = J+r;   /* J(i-1,j  ) */
      p  = d+1;   /* J(i  ,j  ) */
      q  = I+r+1; /* I(i  ,j  ) */
      for ( i=X-2; i; --i )
         {
          m = *b<*d ? *b : *d;
          m =  m<*p ?  m : *p;
         *p =  m>*q ?  m : *q;
         ++b;
         ++d;
         ++p;
         ++q;
         }
      }
   return;
   }


void BackScanBot4( I, J, X, Y )
   byte *I;
   byte *J;
   int X,Y;
   {
   byte *b,*d,*p,*q;
   byte m;
   int i,j;
   int r,s;
   
   r = (X*Y)-1;
   for ( j=Y-1; j; --j )
      {
      s  = r;     /* pixel (i+1,j+1) loc */
      r -= X;     /* pixel above (i+1,j) loc */
      b  = J+s-1; /* J(i  ,j+1) */
      d  = J+r;   /* J(i+1,j  ) */ 
      p  = d-1;   /* J(i  ,j  ) */
      q  = I+r-1; /* I(i  ,j  ) */
      for ( i=X-2; i; --i )
         {
          m = *b<*d ? *b : *d;
          m =  m<*p ?  m : *p;
         *p =  m>*q ?  m : *q;
         --b;
         --d;
         --p;
         --q;
         }
      }
   return;
   }






int Difference( J, J2, N )
   byte *J,*J2;
   int N;
   {
   int i,d;
   byte *u,*v;

   d=0;
   u=J;
   v=J2;
   for (i=N; i; --i )
      {
      d += abs( *(u++) - *(v++) );
      }
   return(d);
   }



void MakeTopBot( I, J, T, B, N)
   byte *I,*J;
   byte *T,*B;
   int N;
   {
   int k;
   byte *orig,*smooth,*top,*bot;

   orig = I;
   smooth = J;
   top = T;
   bot = B;

   for (k=N; k; --k)
      {
      if ( *orig > *smooth )
         {
         *top = *orig;
         *bot = *smooth;
         }
      else if ( *orig < *smooth )
         {
         *bot = *orig;
         *top = *smooth;
         }
      else
         {
         *top = *smooth;
         *bot = *smooth;
         }
      ++orig;
      ++smooth;
      ++top;
      ++bot;
      }
   return;
   }




void MakeRecon( I, J, T, B, N)
   byte *I,*J;
   byte *T,*B;
   int N;
   {
   int k;
   byte *orig,*recon,*top,*bot;

   orig = I;
   recon = J;
   top = T;
   bot = B;

   for (k=N; k; --k)
      {
      if ( (*top == *orig) || (*bot == *orig) )
         {
         *recon = *orig;
         }
      else if ( *top <  *orig )
         {
         *recon = *top;
         }
      else
         {
         *recon = *bot;
         }
      ++orig;
      ++recon;
      ++top;
      ++bot;
      }
   return;
   }
