/* reconspec - compute the reconstructed opening spectrum of an image  */

/* dependencies: ReadRasterFile(), MorphSub(), WriteRasterFile(), ReconSub() */

/* reconspec version 1.0   24 February 1994         */
/*                                                 */
/* by Richard Alan Peters II                       */
/* Department of Electrical Engineering            */
/* Vanderbilt University School of Engineering     */
/* Nashville, TN 37235                             */
/* rap2@vuse.vanderbilt.edu                        */
/*                                                 */
/* This software is freely redistributable if      */
/* the author's name and affiliation are included. */

#include "morph_sub.h"
#include "rasterio.h"
#include "recon.h"
#include "reconspec.h"


main( argc, argv ) 
   unsigned int argc;    /* count of arguments parsed by operating system    */
   char *argv[];         /* pointers to arguments parsed by operating system */
   {
   struct rasterfile H;    /* raster file header */
   unsigned char *C;       /* colormap */
   byte  *I,*O;            /* original (or tophat) image and open image */
   FILE  *fp;              /* file pointer */
   char  *SElistFname;     /* name of file containing list of SE filenemaes */
   char **SElist;          /* list of SE filenames */
   int    SEdim[NUMBANDS]; /* dimensions of SEs (not file based) */
   int    SEtype;          /* disk, horiz. line, vert. line, def'd. in file */
   int    bands;           /* number of spectral bands to compute */
   int    i,j,k;
   int    M,X,Y;
   int    verbose;
   int    nbhd;
   char   fname[256];
   char   *dot;

   /* begin */



   if (argc < 3 ) /* display usage */
      {
      fprintf(stderr,
         "usage: reconspec  InFile  OutFileBaseName  \n\
                 [-s d|h|v|[f SElistFile]]  [-d n [n n ...n]] [-4|8] [-v]\n");
      exit(0);
      }


   /* initialize everything that needs it */



   /* Get the arguments */

    GetCLArgs(argc,argv,&SEtype,&SElistFname,&bands,SEdim,&nbhd,&verbose);


   /* if we're to use SE files, get the names, and check the contents. */

   if ( SEtype == SEFL )
      {
      CheckSEfiles(SElistFname,&SElist,&bands);
      }

 
   /* talk back if requested */

   if ( verbose )
      {
      fprintf(stderr,"Will compute %d bands:",bands);
      if ( SEtype == SEFL )
         {
         fprintf( stderr,"\n");
         fprintf( stderr,"0: single-pixel SE\n");
         for( i=0; i<bands-1; ++i )
            {
            fprintf( stderr,"%d: %s\n",i+1,*(SElist+i));
            }
         }
      else 
         {
         for ( i=0; i<bands; ++i)
            fprintf(stderr," %d",*(SEdim+i));
         fprintf(stderr,".\n");
         }
      }


   /* read the input image into memory */

   if ( !(fp = OpenFile( argv[1], "IMGPATHI", "r" )) ) exit( 0 );
   ReadRasterFile( fp, &H, &C, &I, 0, 0, 0, 0, ALLOC );
   fclose( fp );


   /* extract the luminance */

   ExtractLuminance( &H, C, I, 0, 0, 0, 0 );


   /* get the image dimensions */

   X = H.ras_width;
   Y = H.ras_height;
   M = X*Y;


   /* allocate image space */

   if ( !( O = (byte *)calloc( M, sizeof(byte) )) )
      {
      fprintf(stderr,"main: Unable to allocate image buffer.\n");
      exit(0);
      } 

   /* setup output file name */

   strcpy( fname, argv[2] );
   dot = rindex( fname, '.' );
   if ( !dot || (dot == fname) ) 
      dot = fname + strlen( fname );



   /* main loop */

   for ( i=bands-1; i>0; --i )
      {
      if ( verbose )
         {
         fprintf( stderr,"band: %d  ",i);
         fflush( stderr );
         }

/* DumpImage( I, X, Y, "test.ras" ); */

      if ( SEtype == DISK )
         {
         MorphSub( OPEN,I,O,X,Y,GRAIMG,FALSE,0,0,NULL,AUTO,
                   *(SEdim+i),*(SEdim+i),0,GRASE,SET,0,FALSE,FALSE);
         }
      else if ( SEtype == HORZ )
         {
         MorphSub( OPEN,I,O,X,Y,GRAIMG,FALSE,0,0,NULL,AUTO,
                   *(SEdim+i),1,0,GRASE,SET,0,FALSE,FALSE);
         }
      else if ( SEtype == VERT )
         {
         MorphSub( OPEN,I,O,X,Y,GRAIMG,FALSE,0,0,NULL,AUTO,
                   1,*(SEdim+i),0,GRASE,SET,0,FALSE,FALSE);
         }
      else   /* SEtype == SEFL */
         {
         MorphSub( OPEN,I,O,X,Y,GRAIMG,FALSE,0,0,*(SElist+i-1),0,
                   0,0,0,GRASE,SET,0,FALSE,FALSE);
         }


      /* reconstruct residual */
      ReconSub( I, O, O, X, Y, nbhd, verbose );


      /* write the opened image out */
      sprintf( dot, ".%03d", i );
      strcat( fname, ".ras" );
      if ( !(fp = OpenFile( fname, "IMGPATHO", "w" )) ) exit( 0 );
      WriteRasterFile( fp, H, C, O, 0, 0, 0, 0 );
      fflush( fp );
      fclose( fp );
      if ( verbose )
         {
         fprintf( stderr,"%s\n",fname);
         }


      /* compute residual */
      SubImg( I, O, I, X, Y );
      }

   /* first band (band 0) is final residual */


   if ( verbose )
      {
      fprintf( stderr,"band: 0  ");
      fflush( stderr );
      }

   sprintf( dot, ".%03d", i );
   strcat( fname, ".ras" );
   if ( !(fp = OpenFile( fname, "IMGPATHO", "w" )) ) exit( 0 );
   WriteRasterFile( fp, H, C, I, 0, 0, 0, 0 );
   fflush( fp );
   fclose( fp );
   if ( verbose )
      {
      fprintf( stderr,"%s\n",fname);
      }

   free( I );
   free( O );
   if ( C ) free( C );
   if ( SEtype == SEFL ) free( SElist );

   }





/* parse (argc, argv) argument list for morph */

static void  GetCLArgs( argc,argv,SEtype,SElist,bands,SEdim,nbhd,verbose )
   unsigned int argc; /* count of arguments parsed by operating system    */
   char  *argv[];      /* pointers to arguments parsed by operating system */
   int   *SEtype;      /* disk, horizontal line, vertical line, or file.   */
   char **SElist;      /* SE-file list filename */
   int   *bands;       /* number of bands in the spectrum */
   int   *SEdim;       /* size of SEs */
   int   *nbhd;        /* either 4 or 8-connected neighborhood */
   int   *verbose;     /* T => give lots of info back */
   {
   int n = argc;
   int i,j;
   char *s;

   /* set args to default values */
   *SEtype = DISK;
   *SElist = NULL;
   *bands = 6;
   *SEdim     =  1;
   *(SEdim+1) =  3;
   *(SEdim+2) =  5;
   *(SEdim+3) =  9;
   *(SEdim+4) = 17;
   *(SEdim+5) = 33;
   *verbose = FALSE;
   *nbhd = 4;
   s = "disk";

   /* get arguments */
   while ( --n )
      {
      if ( argv[n][0] == '-' )          /* then this is a switch */
         {
         if ( argv[n][1] == 's' )       /* SE type follows */
            {
            if ( argv[n+1][0] == 'd' )
               {
               *SEtype = DISK;
               s = "disk";
               }
            else if ( argv[n+1][0] == 'h' )
               {
               *SEtype = HORZ;
               s = "horizontal line";
               }
            else if ( argv[n+1][0] == 'v' )
               {
               *SEtype = VERT;
               s = "vertical line";
               }
            else if ( argv[n+1][0] == 'f' )
               {
               *SEtype = SEFL;
               *SElist = argv[n+2];
               *bands  = 0;
               *SEdim  = 0;
               s = "file-defined";
               }
              else 
               {
               fprintf( stderr,"-s must be followed by one of d,h,v,f.\n");
               exit(0);
               }
            }
         else if ( argv[n][1] == 'd' ) /* SEdims follow */
            {
            if ( *SEtype == SEFL )
               {
               fprintf( stderr,"Cannot specify dims (-d) with file def'd. SEs.\n");
               exit(0);
               }

            if ( n+1<argc && isdigit( argv[n+1][0] ) ) /* a number follows -d */
               {
               *bands = atoi( argv[n+1] ); /* it is the number of bands */
               if (*bands == 1) 
                  {     /* if there is only one */
                  fprintf( stderr,"One band is original image.\n");
                  exit( 0 );
                  }
               else /* there is more than one band */
                  {
                  i=1;
                  *SEdim = 1;
                  if ( n+2<argc && isdigit( argv[n+2][0] ) ) 
                     {   /* then at least one dim is given */
                     if ( atoi( argv[n+2] ) != 1 )
                        {   
                        fprintf( stderr,"First band dimension must be 1.\n");
                        exit( 0 );
                        }
                     while ( i<=*bands && n+2+i<argc && isdigit( argv[n+2+i][0] ) )
                        {   /* get the sizes that are specified on the cmd line */
                        *(SEdim+i) = atoi( argv[n+2+i] );
                        if ( *(SEdim+i) <= *(SEdim+i-1) )
                           {
                           fprintf(stderr,"SE sizes must increase.\n");
                           exit( 0 );
                           }
                        ++i;
                        }
                     }
                  if ( i<*bands ) /* there are bands that were not specified */
                     {
                     if ( i > 1 ) /* get smallest integer > last SE size */
                        {
                        j = (SEdim[i-1] / 2) + 1;
                        }  
                     else
                        {
                        j = 1; /* no bands were specified, start with 3 */
                        }  
                     while ( i<*bands )
                        {
                        *(SEdim+(i++)) = (2 * (j++)) + 1;
                        }
                     }
                  }
               }
            else 
               {
               fprintf( stderr,"-l must be followed by 1 or more integers.\n");
               exit(0);
               }
            }
         else if ( argv[n][1] == 'v' ) /* turn on verbose flag */
            {
            *verbose = TRUE;
            }
         else if ( argv[n][1] == '4' ) /* turn on verbose flag */
            {
            *nbhd = 4;
            }
         else if ( argv[n][1] == '8' ) /* turn on verbose flag */
            {
            *nbhd = 8;
            }
         else
            {
            fprintf( stderr,"%s is an invalid argument.\n",argv[n]);
            exit(0);
            }
         }    /* end of if ( argv[n][0] == '-' ) */
      }   /* end of  while ( --n ) "get arguments" loop */

   if ( *verbose )
      {
      fprintf( stderr,"Will open using %s structuring elements.\n",s);
      }

   return;
   }



static void CheckSEfiles(SElistFname,SElist,bands)
   char  *SElistFname; /* name of file containing list of SE filenemaes */
   char **SElist;      /* list of SE filenames */
   int   *bands;       /* number of spectral bands to compute */
   {
   FILE  *SElistfp;
   FILE  *SEfp;
   char   SEfname[256];
   int    charcount;
   int    sX,sY,osX,osY;
   int    i;
   
   charcount = 0;
   osX = osY = 0;
   *bands = 0;

   if ( !(SElistfp = OpenFile( SElistFname, "", "r" )) ) exit( 0 );
   fscanf( SElistfp, "%s", SEfname );
   while ( !feof(SElistfp) )
      {
      ++(*bands);
      if ( !(SEfp = OpenFile( SEfname, SEPATH, "r" )) ) exit( 0 );
      charcount += strlen( SEfname ) + 1;
      fscanf( SEfp, "%d %d", &sX, &sY );
      if ( (sX < osX) || (sY < osY) )
         {
         fprintf( stderr,"The SE in %s is smaller than the previous.\n");
         fprintf( stderr,"The SEs must be nondecreasing in size.\n");
         exit(0);
         }
      fclose( SEfp );
      fscanf( SElistfp, "%s", SEfname );
      }

   if ( !(*bands) )
      {
      fprintf( stderr,"No SE files are specified in %s.\n",SElistFname);
      exit(0);
      }

   fclose( SElistfp );
   charcount += *bands * sizeof( char * );
   
   if ( !( *SElist = (char *)calloc( charcount, sizeof(char) )) )
      {
      fprintf(stderr,"Unable to allocate filename list.\n");
      exit(0);
      } 

   charcount = *bands * sizeof( char * );
   if ( !(SElistfp = OpenFile( SElistFname, "", "r" )) ) exit( 0 );
   for ( i=0; i < *bands; ++i )
      {
      fscanf( SElistfp, "%s", SEfname );
      *((char **)*SElist + i) = *SElist + charcount;
      strcpy( *SElist+charcount, SEfname );
      charcount += strlen( SEfname ) + 1;
      }
   fclose( SEfp );

   /* there is always one more band than SE file */
   ++(*bands);

   return;
   }
 

static void SubImg( I1, I2, O, X, Y )
   byte *I1;  /* image 1 */
   byte *I2;  /* image 2 */
   byte *O;   /* result  */
   int X,Y;   /* image horizontal, vertical dimensions */
   {
   byte *A = I1;
   byte *B = I2;
   byte *C = O;
   int   n = X*Y;
   int i;

   for (i=n; i; --i)
      {
      *(C++) = *(A++) - *(B++);
      }

   return;
   }



static char *rindex( s, c)
   char *s;
   char c;
   {
   int n;
   char *t;

   n = strlen( s );
   t = s+n-1;
   while ( n && (*t != c) )
      {
      --t;
      --n;
      }
   if ( n )
      return( t );
   else
      return( NULL );
   }


     
  


