/* avg3df -- performs a moving combination of a floating-point image     */
/*           sequence (raw-format)                                       */
/*                                                                       */
/* neighborhoods: 1x1xn, 3x3xn (2D-plus x n), 3x3xn (square x n), or     */
/*                5x5xn (2D-fat-plus x n)                                */
/*                                                                       */
/* operations: average (avg)                                             */
/*             minimum (min)                                             */
/*             maximum (max)                                             */
/*             median  (med)                                             */
/*             rank or order statistic (os) nnn (rank 1 == max)          */
/*             ordinary linear combination (olc) coefficients a1 ... aN  */
/*             rank-ordered linear comb. (rlc) coefficients a1 ... aN    */


/* version 1.0     11 Aug 1992 */

/* by Richard Alan Peters II                       */
/* Department of Electrical Engineering            */
/* Vanderbilt University School of Engineering     */ 
/* Nashville, TN 37235                             */ 
/* rap2@vuse.vanderbilt.edu                        */ 
/*                                                 */ 
/* This software is freely redistributable if      */ 
/* the author's name and affiliation are included. */


#include <math.h>
#include <memory.h>		
#include <stdio.h>
#include <string.h>
#include <values.h>
#include <sys/types.h>
#include "avg3df.h"


main( argc, argv ) 
   unsigned int argc;    /* count of arguments parsed by operating system    */
   char *argv[];         /* pointers to arguments parsed by operating system */
   {
   FILE *fptr;
   char filename[80];
   char compname[80];
   char *basename;
   char *outbase;
   char *suffix;
   char *slash;
   float **imgptr;
   float *coefflist;
   float *c;
   float f,r,s;
   float *I,*O,*T;	    /* image pointers */
   float *u,*v,*w,*z;   /* image pointers */
   int X,Y,N;           /* image dimensions */
   float *pix;          /* pixel sort array */
   int i,j,k;           /* indices */
   int order;
   int InIndex;
   int OutIndex;
   int frames;
   int ncoeffs;
   int n = argc;
   int mode;
   int rank;
   int done;
   int first;
   int kernel;

   char Index[IXLGTH+2];
   char Format[FTLGTH];



   if (argc < 5)
      {
      fprintf(stderr,
         "usage: avg3d -i basename  h-dim  v-dim [.suffix]\n\
             -o output_basename\n\
             -r nnn  [-s nnn]  [-n nnn]\n\
             -c avg | min | max | med | os nnn | olc[|rlc] a1 ... aN\n\
             -k 1x1 | plus | 3x3 | 5x5 \n");
      exit(0);
      }

   basename = (char *)NULL;
   outbase = (char *)NULL;
   suffix = (char *)NULL;
   frames = -1;
   InIndex = 1;
   kernel = K1X1;
   mode = MAVG;
   rank = 0;
   order = DFLTORD;

   /* get arguments */
   while ( --n )
      {
      if ( argv[n][0] == '-' )          /* then this is a switch */
         {
         if ( argv[n][1] == 'i' )       /* basename and dim's follow */
            {
            basename = argv[n+1];
            if ( (n+2 > argc) || !isdigit(argv[n+2]) || 
                 (n+3 > argc) || !argv[n+3] )
               {
               fprintf(stderr,"Image (frame) dimensions must be specified\n");
               exit( 0 );
               }
            X = atoi(argv[n+2]);
            Y = atoi(argv[n+3]);
            if ( (n+4 < argc) && (argv[n+4][0] == '.') )
               suffix = argv[n+4];
            }
         else if ( argv[n][1] == 'o' )       /* output basename follows */
            {
            outbase = argv[n+1];
            }
         else if ( argv[n][1] == 'r' )       /* order of average follows */
            {
            order = atoi(argv[n+1]);
            }
         else if ( argv[n][1] == 's' )       /* starting index follows */
            {
            InIndex = atoi(argv[n+1]);
            }
         else if ( argv[n][1] == 'n' )       /* number of frames follows */
            {
            frames = atoi(argv[n+1]);
            }
         else if ( argv[n][1] == 'k' )       /* kernel type follows */
            {
            if ( !strcasecmp( argv[n+1], "1x1" ) )
               {
               kernel = K1X1;
               }
            else if ( !strcasecmp( argv[n+1], "3x3" ) )
               {
               kernel = K3X3;
               }
            else if ( !strcasecmp( argv[n+1], "plus" ) )
               {
               kernel = KPLUS;
               }
            else if ( !strcasecmp( argv[n+1], "5x5" ) )
               {
               kernel = K5X5;
               }
            }
         else if ( argv[n][1] == 'c' )       /* combination mode follows */
            {
            if ( !strcasecmp( argv[n+1], "avg" ) )
               {
               mode = MAVG;
               }
            else if ( !strcasecmp( argv[n+1], "min" ) )
               {
               mode = MMIN;
               }
            else if ( !strcasecmp( argv[n+1], "max" ) )
               {
               mode = MMAX;
               }
            else if ( !strcasecmp( argv[n+1], "med" ) )
               {
               mode = MMED;
               }
            else if ( !strcasecmp( argv[n+1], "os" ) )
               {
               mode = MOS;
               rank = atoi( argv[n+2] );
               }
            else if ( !strcasecmp( argv[n+1], "olc" ) )
               {
               mode = MOLC;
               ncoeffs = 0;
               while ( (n+2+ncoeffs<argc) && (argv[n+2+ncoeffs][0] != '-') ) 
                 ++ncoeffs;
               if ( !(coefflist = 
                     (float *)calloc( 1, ncoeffs*(sizeof(float)) )) )
                  {
                  fprintf(stderr,"error allocating coefficient list\n");
                  exit( 0 );
                  }
               for ( i=0; i<ncoeffs; ++i)
                  {
                  *(coefflist+i) = atof(argv[n+2+i]);
                  /*fprintf(stderr,"coeff[%d]: %f\n",i,*(coefflist+i));*/
                  }
               }
            else if ( !strcasecmp( argv[n+1], "rlc" ) )
               {
               mode = MRLC;
               ncoeffs = 0;
               while ( (n+2+ncoeffs<argc) && (argv[n+2+ncoeffs][0] != '-') ) 
                 ++ncoeffs;
               if ( !(coefflist = 
                     (float *)calloc( 1, ncoeffs*(sizeof(float)) )) )
                  {
                  fprintf(stderr,"error allocating coefficient list\n");
                  exit( 0 );
                  }
               for ( i=0; i<ncoeffs; ++i)
                  {
                  *(coefflist+i) = atof(argv[n+2+i]);
                  /*fprintf(stderr,"coeff[%d]: %f\n",i,*(coefflist+i));*/
                  }
               }
            }
         }
      }

   OutIndex = InIndex;

   fprintf(stderr,"Order of combination = %d\n",order);


   /* check args */

   if ( !basename || !outbase )
      {
      fprintf(stderr,"-i and -o must be specified\n");
      exit( 0 );
      }

   if ( !X || !Y )
      {
      fprintf(stderr,"Image (frame) dimensions must be specified (& nonzero)\n");
      exit( 0 );
      }

   N = X*Y;

   if ( ((mode == MOLC) || (mode == MRLC)) && (order != ncoeffs) )
      {
      fprintf(stderr,"Number of coefficients must equal order of combination\n");
      fprintf(stderr,"One coefficient multiplies each interframe nbhd.\n");
      exit( 0 );
      }


   /* check median or order-stats requests */

   if ( mode == MMED )
      {
      rank = order/2+1;
      }
   else if ( mode == MOS )
      {
      if ( (rank < 0) || (rank > order) )
         {
         fprintf(stderr,"Rank must be between 1 and %d\n",order);
         exit( 0 );
         }
      else if ( rank == 1 )
         {
         mode == MMAX;
         }
      else if ( rank == order )
         {
         mode == MMIN;
         }
      }


   /* allocate image pointer list */      
   if ( !(imgptr = (float **)calloc( order, sizeof(float *) )) )
      {
      fprintf(stderr,"error allocating image pointer list\n");
      exit( 0 );
      }


   /* allocate image buffers */

   *imgptr = I;
   for ( i=1; i<order; ++i)
      {
      if ( !(*(imgptr+i) = (float *)calloc( N, sizeof(float)) ) )
         {
         fprintf(stderr,"error allocating image %d\n", i);
         exit( 0 );
         }
      }
   if ( !(O = (float *)calloc( N, sizeof(float)) ) )
      {
      fprintf(stderr,"error allocating output image \n");
      exit( 0 );
      }

   if ( !( (mode == MAVG)  || (mode == MOLC) ) )
      if ( !(pix = (float *)calloc( order, sizeof(float)) ) )
         {
         fprintf(stderr,"error allocating pixel sort array \n");
         exit( 0 );
         }

   /* write index field width in format string */
   sprintf( Format, "%%0%1dd", IXLGTH );
   Index[0] = '.';

   /* main loop */
   done = FALSE;
   first = TRUE;
   while ( frames )
      {
      /* get the next image */

      if ( !first ) /* not the first image */
         {
         T = *imgptr;
         for ( i=0; i<order-1; ++i )
            {
            *(imgptr+i) = *(imgptr+i+1);
            }
         *(imgptr+i) = T;

         /* make the input file name index */
         sprintf( Index+1, Format, InIndex );

         if ( frames - (order>>1) ) /* nonzero => there are more frames */
            {
            strcpy( filename, basename );
            strcat( filename, Index );
            if ( suffix ) 
               strcat( filename, suffix );

            if ( !(fptr = fopen( filename, "r" )) )
               {
               fprintf(stderr,"Unable to open file %s \n",filename);
               done = TRUE;
               }
            else
               {
               if ( fread(imgptr+i, sizeof(float), N, fptr) != N )
                  {
                  fprintf(stderr,"Error reading file %s \n",filename);
                  exit( 0 );
                  }
               fclose(fptr);
               }
            }
         else  /* there are no more frames */
            {
            done = TRUE;
            }

         if ( done ) /* zero this image */
            {
            memset( *(imgptr+i), '\000', N );
            }
         ++InIndex;
         }
      else /* this is the first image */
         {
         for ( i=(order>>1); i<order; ++i) /* read in the leading images */
            {
            /* make the input file name index */
            sprintf( Index+1, Format, InIndex );
            strcpy( filename, basename );
            strcat( filename, Index );
            if ( suffix ) 
               strcat( filename, suffix );
            if ( !(fptr = fopen( filename, "r" )) )
               {
               fprintf(stderr,"Unable to open file %s \n",filename);
               exit(0);
               }
            else
               {
               if ( fread(imgptr+i, sizeof(float), N, fptr) != N )
                  {
                  fprintf(stderr,"Error reading file %s \n",filename);
                  exit( 0 );
                  }
               fclose(fptr);
               }
            ++InIndex;
            }
         first = FALSE;
         }

      /* do the various combinations */
      if ( mode == MAVG)
         {
         z=O;
         if ( kernel == K1X1 )
            {
            for ( j=0; j<N; ++j )
               {
               r = 0;
               for ( i=0; i<order; i++ )
                  r += *(*(imgptr+i)+j);
               *(z++) = r / order;
               }
            }
         }
       else if ( mode == MOLC )  /* ordinary linear combination */
         {
         z=O;
         for ( j=0; j<N; ++j )
            {
            f = 0;
            c = coefflist;
            for ( i=0; i<order; i++ )
               f += *(c++) * (float)(*(*(imgptr+i)+j));
            *(z++) = (byte)(f > WHITE) 
                   ? WHITE : ( (f < BLACK) ? BLACK : f );
            }
         }
      else if ( mode == MMAX )
         {
         z=O;
         for ( j=0; j<N; ++j )
            {
            r = BLACK;
            for ( i=0; i<order; i++ )
               {
               s = *(*(imgptr+i)+j);
               if ( s > r )
                  r = s;
               }
            *(z++) = r;
            }
         }
      else if ( mode == MMIN )
         {
         z=O;
         for ( j=0; j<N; ++j )
            {
            r = WHITE;
            for ( i=0; i<order; i++ )
               {
               s = *(*(imgptr+i)+j);
               if ( s < r )
                  r = s;
               }
            *(z++) = r;
            }
         }
      else if ( (mode == MMED) || (mode == MOS) || (mode == MRLC) )
         {
         z=O;
         for ( j=0; j<N; ++j ) /* process loop */
            {
            w = pix;
            for ( i=0; i<order; i++ )  /* get pixels */
               {
               *(w++) = *(*(imgptr+i)+j);
               }

            for ( i=1; i<order; i++ )  /* sort loop */
               {
               v = pix+i;
               u = --v;
               if ( *u > *v )  /* order loop */
                  {
                   r = *u;
                  *u = *v;
                  *v =  r;
                  j=i-1;
                  while ( j )  /* bubble loop */
                     {
                     v = pix+j;
                     u = --v;
                     if ( *u <= *v ) /* are ok */
                        {
                        break;       /* go to next order */
                        }
                     else            /* swap them and */
                        {
                         r = *u;
                        *u = *v;
                        *v =  r;
                        }
                     --j;            /* check next bubble */
                     }  /* bubble loop */
                  }  /* order loop */
               }  /* sort loop */

            if ( mode != MRLC) /* (mode == MMED) || (mode == MOS) */
               {
               *(z++) = *(pix + order - rank);
               }
            else /* mode == MRLC*/
               {
               f = 0;
               u = pix;
               c = coefflist + ncoeffs;
               for ( i=0; i<order; i++ )
                  {
                  f += *(--c) * (float)*(u++);
                  }
               *(z++) = f;
               }

            }  /* process loop */
         }  /* mode == MMED, MOS, or MRLC*/


      /* make the output file */

            else
            ++InIndex;

      /* make the file name index */
      sprintf( Index+1, Format, OutIndex );
      strcpy( filename, outbase );
      strcat( filename, Index );
      if ( suffix ) 
         strcat( filename, suffix );

      fprintf( stderr, "%s\n", filename );

      if ( !(fptr = fopen( filename, "w" )) )
         {
         fprintf(stderr,"Unable to open output file %s \n",filename);
         exit(0);
         }
      else
         {
         if ( fwrite(O, sizeof(float), N, fptr) != N )
            {
            fprintf(stderr,"Error writing file %s \n",filename);
            exit( 0 );
            }
         fclose(fptr);
         }

      ++OutIndex;
      --frames;
      }
               
   }




