
/* unstuck -- aprogram to remove stuck pixels from image time sequnces. */
/*                                                                      */
/* The program will use either an alpha-trimmed median (ATM) filter or  */
/* a variant of the lower-upper-middle (LUM) filter.                    */
/*                                                                      */
/* The LUM takes an x-by-y-by-t neigborhood (x, y, t, odd) and divides  */
/* it into center and surround (LUM) neighborhoods. The center nbhd is  */
/* the set C = { I( x/2, y/2, k ) | k in [0,...,t-1] }. If the standard */
/* deviation of C is greater than a user specifed tolerance, delta,     */
/* then the center pixel I( x/2, y/2, t/2 ) is passed unchanged to the  */
/* output image. (If there is a sufficiently large variation in the     */
/* pixel value over adjacent frames, the pixel is probably not stuck.)  */
/* The surround or LUM nbhd is the set S = { I( i, j, k ) | i in        */
/* [0,...,x-1], j in [0,...,y-1], k in [0,...,t-1]} \ C  U med(C). That */
/* is S is the entire spacetime nbhd sans the center nbhd, C,  plus the */
/* median value from C. If the std. dev. of C is within tolerance delta */
/* an LUM smoothing filter is applied to S.                             */
/* A variation of the algorithm (specified by the -z flag) defines S as */
/* S = { I( i, j, t/2 ) | i in [0,...,x-1], j in [0,...,y-1] } \        */
/* I( x/2, y/2, t/2 )  U med(C). That is, the LUM is calculated only in */
/* the surround restricted to the current frame.                        */
/*                                                                      */
/* The order of the filter is restricted to lie in [0,...,m-1], where m */
/* is the index of the median of the surround.                          */
/* Note that if order == 0, then the output is the median of the        */
/* surround whenever the std. dev. of the center is less than delta.    */
/* If the order were equal to m, then the filter would pass every pixel */
/* unchanged.                                                           */

/* version 2.0     9 November 1993 */

/* by Richard Alan Peters II                       */
/* Department of Electrical Engineering            */
/* Vanderbilt University School of Engineering     */ 
/* Nashville, TN 37235                             */ 
/* rap2@vuse.vanderbilt.edu                        */ 
/*                                                 */ 
/* This software is freely redistributable if      */ 
/* the author's name and affiliation are included. */


#include <math.h>
#include <memory.h>		
#include <stdio.h>
#include <string.h>
#include <values.h>
#include <sys/types.h>
#include "rasterio.h"
#include "unstuck.h"

static int  median();
static void bsort();
static void lumf();
static float stddev();

main( argc, argv ) 
   unsigned int argc;    /* count of arguments parsed by operating system    */
   char *argv[];         /* pointers to arguments parsed by operating system */
   {

   byte **imgptr;       /* frame sequence buffer (FIFO) pointer */
   struct rasterfile H; /* rasterfile headers */
   byte *cmap;          /* colormaps */
   byte *B,*I,*O,*T;    /* image pointers */
   byte *u,*v,*w,*z;   	/* image pointers */
   int atm;             /* T ==> use alpha trimmed median filter */
   int lum;             /* T ==> use lower-upper-middle filter */
   int  *cent,*csa;     /* center neighborhood sort array */
   int  *srnd,*ssa;     /* surround neighborhood sort array */
   int X,Y,N;           /* image dimensions */
   int x,y,t;           /* neighborhood dimensions */
   int XE,YE;           /* x-end, y-end */
   int i,j,k,l,m,n;     /* indices */
   int r,s,q;           /* coordinates of center of spatial nbhd */
   int cmed,smed;       /* central median value, surround nbhd median value */
   int cmix,smix;       /* indices of center, surround nbhd medians */
   int nos,ros;         /* neighborhood and row offsets */
   int InIndex;         /* input file index  */
   int OutIndex;        /* output file index */
   int CountDown;       /* down counter initiated at end of sequence */
   int frames;          /* number of frames to process (if < 0), no limit */
   int first;           /* index of first image to process */
   int order;           /* order of LUM smoothing filter */
   int upper;           /* upper LUM order stat */
   int lower;           /* lower LUM order stat */
   int support;         /* number of pixels in LUM list */
   int zorder;          /* TRUE ==> use current frame only in surround */
   float csd;           /* std. dev. of center pixel sequence */
   float alpha;         /* tolerence for median difference in ATM filter */
   float delta;         /* tolerance in std.dev. of center pixel sequence */
   float mdiff;         /* difference of medians */

   FILE *fptr;             /* file pointer */
   char filename[80];      /* file name string */
   char *basename;         /* input file base name */
   char *outbase;          /* output file base name */
   char *slash;            /* index of `/` character */
   char Index[IXLGTH+2];   /* file name index character string */
   char Suffix[SXLGTH+1];  /* file name suffix char string */
   char Format[FTLGTH];    /* fprintf format string */



   /* usage info */

   if (argc < 5)
      {
      fprintf(stderr,
         "usage: unstuck -a alpha (atm) | -l order (lum)\n\
           -i input_basename  -o output_basename\n\
           [-d delta (center pixel tolerance; default 0)]\n\
           [-s nnn (index of starting frame; default 1)]\n\
           [-n nnn (number of frames to process; default unlimited)]\n\
           [-x nnn]  [-y nnn]  [-t nnn] (spacetime dims; default 3x3x3)\n\
           [-z] (use zero-order nbhd)\n");
      exit(0);
      }


   /* init variables */

   basename = (char *)NULL;
   outbase = (char *)NULL;
   frames = -1;
   order = -1;
   InIndex = 1;
   delta = 0.0;
   x = y = t = 3;
   n = argc;
   atm = FALSE;
   lum = FALSE;
   zorder = FALSE;


   /* get arguments */

   while ( --n )
      {
      if ( argv[n][0] == '-' )   /* then this is a switch */
         {
         if ( argv[n][1] == 'a' )       /* use an ATM filter  */
            {
            atm = TRUE;
            alpha = atof(argv[n+1]);
            }
         if ( argv[n][1] == 'l' )       /* use an LUM filter  */
            {
            lum = TRUE;
            order = atoi(argv[n+1]);
            }
         if ( argv[n][1] == 'i' )       /* input basename follows */
            {
            basename = argv[n+1];
            }
         else if ( argv[n][1] == 'o' )  /* output basename follows */
            {
            outbase = argv[n+1];
            }
         else if ( argv[n][1] == 's' )  /* starting index follows */
            {
            InIndex = atoi(argv[n+1]);
            }
         else if ( argv[n][1] == 'n' )  /* number of frames follows */
            {
            frames = atoi(argv[n+1]);
            }
         else if ( argv[n][1] == 'r' )  /* order follows */
            {
            order = atoi(argv[n+1]);
            }
         else if ( argv[n][1] == 'd' )  /* delta follows */
            {
            delta = atof(argv[n+1]);
            }
         else if ( argv[n][1] == 'x' )  /* horizontal dimension follows */
            {
            x = atoi(argv[n+1]);
            }
         else if ( argv[n][1] == 'y' )  /* vertical dimension follows */
            {
            y = atoi(argv[n+1]);
            }
         else if ( argv[n][1] == 't' )  /* time duration follows */
            {
            t = atoi(argv[n+1]);
            }
         else if ( argv[n][1] == 'z' )  /* use 0-order surround nbhd */
            {
            zorder = TRUE;
            }
         }
      }

   if ( t <= 1 )
      {
      fprintf( stderr, "Number of frames, t, must be greater than 1.\n");
      exit( 0 );
      }

   /* calculate the support and the index of the median of the LUM filter */

   if ( !zorder )
      support = ((x * y * t) - t) + 1;
   else
      support = x * y;
   smix = support / 2;
   cmix = t / 2;
   if ( atm )
      --support;
   CountDown = ( t == 1 ) ? 1 : cmix;


   /* has one of ATM or LUM been selected & is alpha or order within limits? */

   if ( lum )
      {
      if ( (order < 0)  ||  (order > (smix-1)) )
         {
         fprintf(stderr,"You must specify order (-r) in [0,%d].\n",smix-1);
         exit( 0 );
         }
      }
   else if ( atm )
      {
      if ( (alpha <= 0)  ||  (alpha >= 1.0) )
         {
         fprintf(stderr,"You must specify alpha in (0,1).\n");
         exit( 0 );
         }
      }
   else
      {
      fprintf(stderr,"You must select either -a or -l.\n");
      exit( 0 );
      }

   if (  (delta < 0)  ||  (delta > 1.0))
      {
      fprintf(stderr,"You must specify delta in [0,1].\n");
      exit( 0 );
      }



   /* are base-names specified? */

   if ( !basename || !outbase )
      {
      fprintf(stderr,"-i and -o must be specified.\n");
      exit( 0 );
      }


   /* check nbhd dimensions */

   if ( !(x%2) || !(y%2) )
      {
      fprintf(stderr,"Neighborhood spatial dimensions must be odd.\n");
      exit( 0 );
      }    


   /* I/O file indices match */

   OutIndex = InIndex;


   /* echo parameters for user */

   if ( lum )
      {
      fprintf(stderr,"LUM filter.  Order = %d.\n",order);
      }
   else
      {
      fprintf(stderr,"ATM filter.  Alpha = %f.\n",alpha);
      }
   fprintf(stderr,"neighborhood dimensions: (%d,%d,%d)   ",x,y,t);
   if ( zorder )
      {
      fprintf(stderr,"zero-order (2D) surround requested.\n");
      }
   fprintf(stderr,"delta: %f\n",delta);


   /* write index field width in format string */

   sprintf( Format, "%%0%1dd", IXLGTH );


   /* create index field for output file name */

   Index[0] = '.';
   sprintf( Index+1, Format, InIndex );


   /* make output file name from basename + index + suffix */

   strcpy( filename, basename);
   strcat( filename, Index );
   strcat( filename, ".ras" );


   /* open the first file. (This also allocates spce for the first image.) */

   if ( !(fptr = OpenFile( filename, "", "r" )) ) exit( 0 );
   if (  ReadRasterFile( fptr, &H, &cmap, &I, 0, 0, 0, 0, ALLOC )  ) exit( 0 );
   fclose( fptr );


   /* get size info */

   X = H.ras_width;
   Y = H.ras_height;
   N = X*Y;


   /* allocate frame FIFO and image buffers */

   if ( !(imgptr = (byte **)calloc( t, sizeof(byte *) )) )
      {
      fprintf(stderr,"Error allocating image pointer list.\n");
      exit( 0 );
      }

   *imgptr = I;            /* 1st image buffer has already been allocated */
   memset( I, '\000', N ); /* zero it out */

   for ( i=1; i<t; ++i) /* now allocate and assign the others */
      {
      if ( !(*(imgptr+i) = (byte *)calloc( N, sizeof(byte)) ) )
         {
         fprintf(stderr,"error allocating image %d\n", i);
         exit( 0 );
         }
      }
   if ( !(O = (byte *)calloc( N, sizeof(byte)) ) ) /* allocate output image */
      {
      fprintf(stderr,"error allocating output image \n");
      exit( 0 );
      }


   /* compute control loop index extrema */
   /* note: this program does not change any pixels in the border area */

   XE = X-(x-1); /* index of last center pixel in row */
   YE = Y-(y-1); /* index of last row of center pixels in image */


   /* allocate sort buffers */

   if ( !(cent = (int *)calloc( t, sizeof(int)) ) )
      {
      fprintf(stderr,"error allocating t-sort buffer \n");
      exit( 0 );
      }

   if ( !(srnd = (int *)calloc( support, sizeof(int)) ) )
      {
      fprintf(stderr,"error allocating surround nbhd-sort buffer \n");
      exit( 0 );
      }


   /* (r,s,q) == indices of center pixel in spacetime nbhd */

   r  = x / 2; 
   s  = y / 2;
   q  = t / 2;


   /* main loop */
   first = TRUE;
   while ( frames  &&  CountDown  )
      {
      /* get the next image */
      if ( !first )  /* then we're into the movie */
         {
         /* scroll the images in the frame FIFO */
         T = *imgptr;
         for ( i=0; i<t-1; ++i )
            {
            *(imgptr+i) = *(imgptr+i+1);
            }
         *(imgptr+i) = T;

         /* make the input file name index */
         sprintf( Index+1, Format, InIndex );

         /* make the input file name */
         strcpy( filename, basename );
         strcat( filename, Index );
         strcat( filename, ".ras" );

         /* try to open the file. */
         if ( fptr = OpenFile( filename, "", "r" ) )
            {
            if (  ReadRasterFile( fptr,&H,&cmap,imgptr+i,0,0,0,0,0 )  ) 
               {
               fprintf(stderr,"error reading input image \n");
               exit( 0 );
               }
            else
               {
               fclose(fptr);
               /* remap the image */
               ExtractLuminance( &H, cmap, *(imgptr+i), 0, 0, 0, 0 );
               }
            }
         else  /* the file doesn't exist */
            {
            memset( *(imgptr+i), '\000', N );
            --CountDown;
            }
         ++InIndex;
         }
      else  /* movie has just started, load up the FIFO */
         {
         for ( i=(t/2); i<t; ++i)
            {
            /* make the input file name index */
            sprintf( Index+1, Format, InIndex );
            strcpy( filename, basename );
            strcat( filename, Index );
            strcat( filename, ".ras" );
            if ( fptr = OpenFile( filename, "", "r" ) )
               {
               if (  ReadRasterFile( fptr,&H,&cmap,imgptr+i,0,0,0,0,0 )  ) 
                  {
                  fprintf(stderr,"error reading input image \n");
                  exit( 0 );
                  }
               /* remap the image */
               ExtractLuminance( &H, cmap, *(imgptr+i), 0, 0, 0, 0 );
               }
             else  /* the file doesn't exist */
               {
               memset( *(imgptr+i), '\000', N );
               --CountDown;
               }
            ++InIndex;
            }
         first = FALSE;
         }

      /* main processing loop */
      for (j=0; j<YE; ++j)                      /*** image row scan */
         {
         ros = j*X;             /* row offset -- # of pixels to row start */
         nos = ros + s*X + r;   /* spatial offset to center of nbhd */
         w = *(imgptr+q) + nos; /* central input pixel */
         z = O + nos;           /* output pixel */ 
     
         for (i=0; i<XE; ++i)                   /*** image column scan */
            {
            csa = cent;  /* central sort array */

            for (k=0; k<t; ++k)                 /*** get center */
               {
               *(csa++) = *(*(imgptr+k) + nos + i);
               }

            csd  = stddev( cent, t ); /* std. dev. of center */
            csd /= (float)WHITE;

            if ( csd > delta ) /* then pass input to output and do next */
               {
               *(z++) = *(w++);
               }

            else  /* this is a stuck pixel */
               {
               n = ros + i; /* offset to begining of neighborhood */
               ssa = srnd;  /* surround sort array */

               if ( zorder )  /* collect zero-order (2D) surround */
                  {
                  u = *(imgptr+q) + n; /* start of 2D neighborhood */

                  for (m=0; m<y; ++m)          /*** neighbohood row scan */
                     {
                     v = u;            /* start of nbhd row */

                     for (l=0; l<x; ++l)       /*** neighbohood column scan */
                        {
                        if ( (m!=r) || (l!=s) ) /* in surround */
                           *(ssa++) = *(v++);  
                        else                    /* in center */
                           ++v;
                        }                      /*** neighbohood column scan */
                     u += X;
                     }                         /*** neighbohood row scan */
                  }      /* end: 2D surround */

               else  /* collect full 3D surround */
                  {
                  for (k=0; k<t; ++k)          /*** neighborhood time scan */
                     {
                     u = *(imgptr+k) + n; /* start of time k spatial nbhd */

                     for (m=0; m<y; ++m)       /*** neighbohood row scan */
                        {
                        v = u;            /* start of nbhd row */
  
                        for (l=0; l<x; ++l)    /*** neighbohood column scan */
                           {
                           if ( (m!=r) || (l!=s) ) /* in surround */
                              *(ssa++) = *(v++);
                           else                    /* in center */
                              ++v;  
                           }
                        u += X;
                        }                      /*** neighbohood row scan */
                     }                         /*** neighborhood time scan */
                  }      /* end: 3D surround */

               /* compute median of center nbhd */
               cmed = median( cent, t );

               if ( lum )
                  {
                  *(srnd + support - 1) = cmed; /* attach c-med to surround */

                  /* compute LUM stats */
                  lumf( srnd, support, order, &upper, &lower);

                  if ( order == 0 )  /* then lower == upper == med(srnd) */
                     {
                     *(z++) = lower;
                     ++w;
                     }
                  else if ( cmed <= lower )
                     {
                     *(z++) = lower;
                     ++w;
                     }
                  else if ( cmed >= upper )
                     {
                     *(z++) = upper;
                     ++w;
                     }
                  else 
                     {
                     *(z++) = *(w++);
                     }
                  }  /* end: LUM */
               else /* is ATM filter */
                  {
                  smed = median( srnd, support);
                  mdiff = (float)abs(cmed-smed) / (float)WHITE;
            
                  if (  mdiff > alpha )
                     {
                     *(z++) = smed;
                     ++w;
                     }
                  else 
                     {
                     *(z++) = *(w++);
                     }
                  }  /* end: ATM */

               }  /* end: this is a stuck pixel */

            }                                   /*** image column scan */
         }                                      /*** image row scan */


      /* make the output file */

      /* make the file name index */
      sprintf( Index+1, Format, OutIndex );
      strcpy( filename, outbase );
      strcat( filename, Index );
      strcat( filename, ".ras" );
      if ( !(fptr = OpenFile( filename, "", "w" )) ) exit( 0 );
      WriteRasterFile( fptr, &H, cmap, O, 0, 0, 0, 0 ); 
      fclose( fptr );
      fprintf( stderr, "%s\n", filename );

      ++OutIndex;
      --frames;
      }   /* end: main loop */
               
   }  /* end: main */



/* compute the standard deviation of the list */

static float stddev( list, n )
   int *list;
   int n;
   {
   int i;
   int *l;
   float p,avg,var;

   avg = 0.0;
   l = list;
   for ( i=n; i; --i )
      {
      avg += (float)(*(l++));
      }
   avg /= (float)n;

   var = 0.0;
   l = list;
   for ( i=n; i; --i )
      {
      p = (float)(*(l++)) - avg;
      var += p * p;
      }
   var = sqrt( var/(float)n );

   return(var);
   }


/* return the median of list */

static int median( list, n )
   int *list;
   int n;
   {
   bsort( list, n );  /* sorts list into descending order */
   return( *(list+(n/2)) );
   }




/* return the upper and lower order stats */

static void lumf( list, n, order, upper, lower)
   int *list;
   int  n;
   int  order;
   int *upper;
   int *lower;
   {
   int med;

   bsort( list, n );  /* sorts list into descending order */
   *upper = *(list+(n/2)-order);
   *lower = *(list+(n/2)+order);
   return;
   }




/* bubble sort list A of N integers into descending order */

static void bsort( A, N )
   int *A;
   int N;
   {
   int i, j;
   int r;
   int *u,*v;

   for ( i=1; i<N; i++ )  /* sort loop */
      {
      v = A+i;
      u = v-1;
      if ( *u < *v )  /* order loop */
         {
         r = *u;
         *u = *v;
         *v =  r;
         j=i-1;
         while ( j )  /* bubble loop */
            {
            v = A+j;
            u = v-1;
            if ( *u >= *v ) /* are ok */
               {
               break;       /* go to next order */
               }
            else            /* swap them and */
               {
                r = *u;
               *u = *v;
               *v =  r;
               }
            --j;            /* check next bubble */
            }  /* bubble loop */
         }  /* order loop */
      }  /* sort loop */

   return;
   }
